"""Define meta information about mapry package."""

__title__ = 'icontract'
__description__ = (
    'Provide design-by-contract with informative violation messages.')
__url__ = 'https://github.com/Parquery/icontract'
__version__ = '2.3.5'
__author__ = 'Marko Ristin'
__author_email__ = 'marko.ristin@gmail.com'
__license__ = 'MIT'
__copyright__ = 'Copyright 2019 Parquery AG'
__status__ = 'Production'
